--
-- Real Shovel for FS19
-- @author:    	kenny456 (kenny456@seznam.cz)
-- @history:	v1.0 - 2020-04-01 - first release
--
RealShovel = {};
RealShovel.confDir = getUserProfileAppPath().. "modsSettings/RealShovel/";
RealShovel.modDirectory = g_currentModDirectory

function RealShovel.prerequisitesPresent(specializations)
	return true
end


function RealShovel.registerOverwrittenFunctions(vehicleType)
end

function RealShovel.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "saveToXmlRealShovel", 						RealShovel.saveToXmlRealShovel)
	SpecializationUtil.registerFunction(vehicleType, "loadFromXmlRealShovel", 						RealShovel.loadFromXmlRealShovel)
	SpecializationUtil.registerFunction(vehicleType, "saveToXmlRealShovelVehicle",					RealShovel.saveToXmlRealShovelVehicle)
	SpecializationUtil.registerFunction(vehicleType, "loadFromXmlRealShovelVehicle",				RealShovel.loadFromXmlRealShovelVehicle)
	SpecializationUtil.registerFunction(vehicleType, "getIsLevelerPickupNodeActive",				RealShovel.getIsLevelerPickupNodeActive)
	SpecializationUtil.registerFunction(vehicleType, "loadLevelerNodeFromXML",						RealShovel.loadLevelerNodeFromXML)
	SpecializationUtil.registerFunction(vehicleType, "updateHeapEffect",							RealShovel.updateHeapEffect)
	SpecializationUtil.registerFunction(vehicleType, "loadDefaultValues",							RealShovel.loadDefaultValues)
	SpecializationUtil.registerFunction(vehicleType, "getConfigFileName",							RealShovel.getConfigFileName)
	SpecializationUtil.registerFunction(vehicleType, "activateModRS",								RealShovel.activateModRS)
end

function RealShovel:registerActionEventsPlayer()
end

function RealShovel:onRegisterActionEvents(isSelected, isOnActiveVehicle)
	local spec = self.spec_realShovel
	if not spec.modInitialized then
		return
	end
	
	if self:getIsActive() then
		self:loadFromXmlRealShovel()
		self:loadFromXmlRealShovelVehicle()
		self:updateHeapEffect(nil, nil, true)
	end
	if g_dedicatedServerInfo ~= nil then
		return
	end
	if spec.event_IDs == nil then
		spec.event_IDs = {}
	end
	if self:getIsActiveForInput() then
		local actions_RC1 = { InputAction.REALSHOVEL_MENU_PLUS, InputAction.REALSHOVEL_MENU_MINUS, InputAction.REALSHOVEL_MENU_CHANGE_DOWN, InputAction.REALSHOVEL_MENU_CHANGE_UP, InputAction.REALSHOVEL_MENU_TOGGLE,
							InputAction.REALSHOVEL_TOGGLE_HELP,  InputAction.REALSHOVEL_TOGGLE }

		for _,actionName in pairs(actions_RC1) do
			local always = (actionName == InputAction.REALSHOVEL_MENU_PLUS or actionName == InputAction.REALSHOVEL_MENU_MINUS) and true or false
			local _, eventID = g_inputBinding:registerActionEvent(actionName, self, RealShovel.actionCallback, true, true, always, true)
			spec.event_IDs[actionName] = eventID
			if g_inputBinding ~= nil and g_inputBinding.events ~= nil and g_inputBinding.events[eventID] ~= nil then
				if actionName == 'something with lower priority' then
					g_inputBinding:setActionEventTextPriority(eventID, GS_PRIO_NORMAL)
				else
					g_inputBinding:setActionEventTextPriority(eventID, GS_PRIO_VERY_HIGH)
				end
				if actionName == InputAction.REALSHOVEL_TOGGLE_HELP then
					g_inputBinding:setActionEventTextVisibility(eventID, true)
				elseif actionName == InputAction.REALSHOVEL_TOGGLE then
					g_inputBinding:setActionEventTextVisibility(eventID, false)
				else
					g_inputBinding:setActionEventTextVisibility(eventID, RealShovel.showHelp)
				end
			end
			local colliding = false
			_, colliding, _ = g_inputBinding:checkEventCollision(actionName)
			if colliding then
				print(string.format("Error: RealShovel got a colliding action: %s", actionName))
			end
		end
	end
end
function RealShovel.registerEventListeners(vehicleType)
	for _,n in pairs( { "onLoad","onPostLoad","onDelete","saveToXMLFile","onUpdate","onDraw","onRegisterActionEvents","registerActionEventsPlayer","saveToXmlRealShovel","loadFromXmlRealShovel","onReadStream","onWriteStream" } ) do
		SpecializationUtil.registerEventListener(vehicleType, n, RealShovel)
	end
end
function RealShovel:getConfigFileName()
	local name = self.configFileName
	local index = name:match'^.*()/'
	local str = string.sub(name, index+1, string.len(name))
	return str
end
function RealShovel:onLoad(savegame)
	self.spec_realShovel = {}
	local spec = self.spec_realShovel
	
	spec.modInitialized = false
	if self.spec_shovel.shovelNodes == nil or #self.spec_shovel.shovelNodes == 0 then
		print("Error: RealShovel initialization failed for "..tostring(self:getFullName()).." !")
		return
	else
		spec.modInitialized = true
	end
	
	local xmlFile = self.xmlFile
	spec.event_IDs = {}
	spec.settingHud = createImageOverlay(RealShovel.modDirectory .. "images\\hud.dds")
	spec.fillUnitIndex = #self.spec_fillUnit.fillUnits + 1
	spec.node = '0|0'
	spec.foldMinLimit = 0
	spec.foldMaxLimit = 1
	spec.pickUpDirection = 1
	spec.effectNode = '0|1|0'
	spec.effectClass = 'LevelerEffect'
	spec.materialType = 'leveler'
	spec.materialTypeId = 1
	spec.fadeTime = 0.5
	spec.speed = -1
	spec.maxHeight = 1.167
	spec.allowFoldingWhileFilled = false
	spec.unit = '$l10n_unit_literShort'
	spec.fillTypeCategories = 'bulk windrow'
	spec.showInShop = false
	spec.timerAdvance = 0
	spec.timerSettings = 0
	spec.selectedValue = 1
	spec.hudText1 = ""
	spec.hudText2 = ""
	spec.hudText3 = ""
	spec.showHud = false
	spec.firstRun = false
	
	self:loadDefaultValues(false, true)
	
	spec.ui = g_gameSettings.uiScale
	spec.sizeHud = spec.ui
	RealShovel.hudPositionX = 0.85
	RealShovel.hudPositionY = 0.4
	RealShovel.showHelp = true
	
	spec.heapI3d = g_i3DManager:loadSharedI3DFile('heap.i3d', RealShovel.modDirectory, false, false)
	spec.effectNodeObject = I3DUtil.indexToObject(spec.heapI3d, '0|1')
	
	local configFileVehicle = RealShovel.confDir .. self:getFullName().."_"..self:getConfigFileName().."/leveler.xml"
	if fileExists(configFileVehicle) then
		self:loadFromXmlRealShovelVehicle(self:getFullName())
		spec.firstRun = false
	else
		createFolder(getUserProfileAppPath().. "modsSettings/")
		createFolder(RealShovel.confDir)
		createFolder(RealShovel.confDir.."/"..self:getFullName().."_"..self:getConfigFileName().."/")
		spec.configXmlVehicle = createXMLFile("RealShovel_XML_Vehicle", configFileVehicle, "vehicle")
		self:saveToXmlRealShovelVehicle(true)
		spec.firstRun = true
	end
	
	if self.spec_fillUnit.fillUnits[spec.fillUnitIndex] ~= nil then
		self.spec_fillUnit.fillUnits[spec.fillUnitIndex] = nil
	end
	local fillUnitConfigurationId = Utils.getNoNil(self.configurations["fillUnit"], 1)
	local baseKey = string.format("vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration(%d).fillUnits", fillUnitConfigurationId -1)
	local i=0
	while true do
		local key = string.format("%s.fillUnit(%d)", baseKey, i)
		if not hasXMLProperty(spec.configXmlVehicle, key) then
			break
		end
		local entry = {}
		if self:loadFillUnitFromXML(spec.configXmlVehicle, key, entry, i+1) then
			table.insert(self.spec_fillUnit.fillUnits, entry)
		else
			g_logManager:xmlWarning(self.configFileName, "Could not load fillUnit for '%s'", key)
			self:setLoadingState(BaseMission.VEHICLE_LOAD_ERROR)
			break
		end
		i = i + 1
	end
	
	spec.nodes = {}
	local levelerNode = {}
	local i = 0
	while true do
		local key = string.format("vehicle.leveler.levelerNode(%d)", i)
		if not hasXMLProperty(spec.configXmlVehicle, key) then
			break
		end
		local levelerNode = {}
		if self:loadLevelerNodeFromXML(levelerNode, spec.configXmlVehicle, key) then
			table.insert(spec.nodes, levelerNode)
		end
		i = i + 1
	end
	spec.pickUpDirection = Utils.getNoNil(getXMLInt(spec.configXmlVehicle, "vehicle.leveler.pickUpDirection"), 1.0)
	spec.litersToPickup = 0
	if not self:getFillUnitExists(spec.fillUnitIndex) then
		g_logManager:xmlWarning(self.configFileName, "Unknown fillUnitIndex '%s' for leveler", tostring(spec.fillUnitIndex))
	end
	
	if self.isClient then
		if spec.effects ~= nil then
			g_effectManager:deleteEffects(spec.effects)
		end
		spec.effects = g_effectManager:loadEffect(spec.configXmlVehicle, "vehicle.leveler.effects", spec.heapI3d, self, self.i3dMappings)
	end
	if self.spec_shovel.shovelNodes ~= nil then
		for i = 1, #self.spec_shovel.shovelNodes do
			if self.spec_shovel.shovelNodes[i].node ~= nil then
				local x,y,z = getWorldTranslation(self.spec_shovel.shovelNodes[i].node)
				setTranslation(spec.effectNodeObject, spec.menuValues['posX'], spec.menuValues['posY'], spec.menuValues['posZ'])
				setScale(spec.effectNodeObject, spec.menuValues['scaleX'], spec.menuValues['scaleY'], spec.menuValues['scaleZ'])
				setTranslation(spec.nodes[1].node, spec.menuValues['nodeX'],spec.menuValues['nodeY'],spec.menuValues['nodeZ'])
				link(self.spec_shovel.shovelNodes[i].node, spec.heapI3d)
				self:updateHeapEffect(nil, nil, true)
			end
		end
		
	end
end
function RealShovel:onPostLoad(savegame)
	local spec = self.spec_realShovel
	if not spec.modInitialized then
		return
	end
	
	local xmlFile = self.xmlFile
	if spec.event_IDs == nil then
		spec.event_IDs = {}
	end
	if self.isClient and g_dedicatedServerInfo == nil then
		local configFile = RealShovel.confDir .. "RealShovelConfig.xml"
		if fileExists(configFile) then
			self:loadFromXmlRealShovel()
		else
			createFolder(getUserProfileAppPath().. "modsSettings/")
			createFolder(RealShovel.confDir)
			RealShovel.configXml = createXMLFile("RealShovel_XML", configFile, "RealShovelConfig")
			self:saveToXmlRealShovel()
		end
	end
	spec.hudWidth = 0.128 * spec.sizeHud
	spec.hudHeight = 0.43 * spec.sizeHud
	spec.hudText1X = RealShovel.hudPositionX + (0.005 * spec.sizeHud)
	spec.hudText2X = spec.hudText1X + (0.01 * spec.sizeHud)
	spec.hudText3X = spec.hudText2X + (0.085 * spec.sizeHud)
	spec.hudTextY = RealShovel.hudPositionY + (0.4 * spec.sizeHud)
end
function RealShovel:onDelete()
	local spec = self.spec_realShovel
	if not spec.modInitialized then
		return
	end
	
    if self.isClient then
		g_i3DManager:releaseSharedI3DFile('heap.i3d', nil, true)
		if spec.effects ~= nil then
			g_effectManager:deleteEffects(spec.effects)
		end
    end
end
function RealShovel:loadLevelerNodeFromXML(levelerNode, xmlFile, key)
	local spec = self.spec_realShovel
	levelerNode.node = I3DUtil.indexToObject(spec.heapI3d, getXMLString(xmlFile, key .. "#node"), self.i3dMappings)
    if levelerNode.node ~= nil then
		levelerNode.width = getXMLFloat(xmlFile, key .. "#width")
        levelerNode.minDropWidth = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#minDropWidth"), levelerNode.width*0.5)
        levelerNode.maxDropWidth = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#maxDropWidth"), levelerNode.width)
        levelerNode.minDropHeight = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#minDropHeight"), 0)
        levelerNode.maxDropHeight = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#maxDropHeight"), 1)
        levelerNode.minDropDirOffset = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#minDropDirOffset"), 0.7)
        levelerNode.maxDropDirOffset = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#maxDropDirOffset"), 0.7)
        levelerNode.numHeightLimitChecks = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#numHeightLimitChecks"), 6)
        levelerNode.lineOffsetPickUp = nil
        levelerNode.lineOffsetDrop = nil
        levelerNode.lastPickUp = 0
        levelerNode.lastDrop = 0
        return true
    end
    return false
end
function RealShovel:getIsLevelerPickupNodeActive(levelerNode)
	local spec = self.spec_realShovel
	
	local fillLevel = self:getFillUnitFillLevel(self.spec_shovel.shovelNodes[1].fillUnitIndex)
	local fillPercentage = fillLevel / self:getFillUnitCapacity(self.spec_shovel.shovelNodes[1].fillUnitIndex)
	if fillPercentage >= 0.9 then
		return spec.menuValues['modActive']
	else
		return false
	end
end
function RealShovel:saveToXMLFile(xmlFile, key)
	local spec = self.spec_realShovel
end
function RealShovel:loadFromXmlRealShovel()
	local spec = self.spec_realShovel
	
	local configFile = RealShovel.confDir .. "RealShovelConfig.xml";
    if self.isClient and g_dedicatedServerInfo == nil and fileExists(configFile) then
		RealShovel.configXml = loadXMLFile("RealShovel_XML", configFile);
		if getXMLBool(RealShovel.configXml, "RealShovelConfig.showHelp") ~= nil then
			RealShovel.showHelp = getXMLBool(RealShovel.configXml, "RealShovelConfig.showHelp")
		end
		if getXMLFloat(RealShovel.configXml, "RealShovelConfig.positionX") ~= nil then
			RealShovel.hudPositionX = getXMLFloat(RealShovel.configXml, "RealShovelConfig.positionX")
		end
		if getXMLFloat(RealShovel.configXml, "RealShovelConfig.positionY") ~= nil then
			RealShovel.hudPositionY = getXMLFloat(RealShovel.configXml, "RealShovelConfig.positionY")
		end
	end
end
function RealShovel:saveToXmlRealShovel(vehicle, active)
	local spec = self.spec_realShovel
    
	if self.isClient and RealShovel.configXml ~= nil and g_dedicatedServerInfo == nil then
		setXMLBool(RealShovel.configXml, "RealShovelConfig.showHelp", RealShovel.showHelp)
		setXMLFloat(RealShovel.configXml, "RealShovelConfig.positionX", RealShovel.hudPositionX)
		setXMLFloat(RealShovel.configXml, "RealShovelConfig.positionY", RealShovel.hudPositionY)
		saveXMLFile(RealShovel.configXml)
	end
end
function RealShovel:loadFromXmlRealShovelVehicle(vehicle, active)
	local spec = self.spec_realShovel
	
	local configFileVehicle = RealShovel.confDir .. self:getFullName().."_"..self:getConfigFileName().."/leveler.xml"
	if  fileExists(configFileVehicle) then
		spec.configXmlVehicle = loadXMLFile("RealShovel_XML_Vehicle", configFileVehicle);
		spec.fillUnitIndex = getXMLInt(spec.configXmlVehicle, "vehicle.leveler#fillUnitIndex")
		spec.node = getXMLString(spec.configXmlVehicle, "vehicle.leveler.levelerNode#node")
		spec.menuValues['width'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.levelerNode#width")
		spec.menuValues['minDropWidth'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.levelerNode#minDropWidth")
		spec.menuValues['maxDropWidth'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.levelerNode#maxDropWidth")
		spec.menuValues['minDropHeight'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.levelerNode#minDropHeight")
		spec.menuValues['maxDropHeight'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.levelerNode#maxDropHeight")
		spec.menuValues['minDropDirOffset'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.levelerNode#minDropDirOffset")
		spec.menuValues['maxDropDirOffset'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.levelerNode#maxDropDirOffset")
		spec.foldMinLimit = getXMLInt(spec.configXmlVehicle, "vehicle.leveler.levelerNode#foldMinLimit")
		spec.foldMaxLimit = getXMLInt(spec.configXmlVehicle, "vehicle.leveler.levelerNode#foldMaxLimit")
		spec.menuValues['nodeX'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.levelerNode#posX")
		spec.menuValues['nodeY'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.levelerNode#posY")
		spec.menuValues['nodeZ'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.levelerNode#posZ")
		spec.menuValues['pickUpDirection'] = getXMLInt(spec.configXmlVehicle, "vehicle.leveler.pickUpDirection")
		spec.effectNode = getXMLString(spec.configXmlVehicle, "vehicle.leveler.effects.effectNode#effectNode")
		spec.effectClass = getXMLString(spec.configXmlVehicle, "vehicle.leveler.effects.effectNode#effectClass")
		spec.materialType = getXMLString(spec.configXmlVehicle, "vehicle.leveler.effects.effectNode#materialType")
		spec.fadeTime = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.effects.effectNode#fadeTime")
		spec.speed = getXMLInt(spec.configXmlVehicle, "vehicle.leveler.effects.effectNode#speed")
		spec.maxHeight = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.effects.effectNode#maxHeight")
		spec.menuValues['posX'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.effects.effectNode#posX")
		spec.menuValues['posY'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.effects.effectNode#posY")
		spec.menuValues['posZ'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.effects.effectNode#posZ")
		spec.menuValues['scaleX'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.effects.effectNode#scaleX")
		spec.menuValues['scaleY'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.effects.effectNode#scaleY")
		spec.menuValues['scaleZ'] = getXMLFloat(spec.configXmlVehicle, "vehicle.leveler.effects.effectNode#scaleZ")
		
		spec.allowFoldingWhileFilled = getXMLBool(spec.configXmlVehicle, "vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration.fillUnits#allowFoldingWhileFilled")
		spec.unit = getXMLString(spec.configXmlVehicle, "vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration.fillUnits.fillUnit#unit")
		spec.fillTypeCategories = getXMLString(spec.configXmlVehicle, "vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration.fillUnits.fillUnit#fillTypeCategories")
		spec.menuValues['capacity'] = getXMLFloat(spec.configXmlVehicle, "vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration.fillUnits.fillUnit#capacity")
		spec.showInShop = getXMLBool(spec.configXmlVehicle, "vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration.fillUnits.fillUnit#showInShop")
		
		spec.menuValues['modActive'] = getXMLBool(spec.configXmlVehicle, "vehicle.leveler#active")
	end
end
function RealShovel:saveToXmlRealShovelVehicle(reload)
	local spec = self.spec_realShovel
    
	if spec.configXmlVehicle ~= nil then
		setXMLString(spec.configXmlVehicle, 'vehicle#type', 'leveler')
		setXMLInt(spec.configXmlVehicle, 'vehicle.leveler#fillUnitIndex', spec.fillUnitIndex)
		setXMLString(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#node', spec.node)
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#width', spec.menuValues['width'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#minDropWidth', spec.menuValues['minDropWidth'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#maxDropWidth', spec.menuValues['maxDropWidth'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#minDropHeight', spec.menuValues['minDropHeight'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#maxDropHeight', spec.menuValues['maxDropHeight'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#minDropDirOffset', spec.menuValues['minDropDirOffset'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#maxDropDirOffset', spec.menuValues['maxDropDirOffset'])
		setXMLInt(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#foldMinLimit', spec.foldMinLimit)
		setXMLInt(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#foldMaxLimit', spec.foldMaxLimit)
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#posX', spec.menuValues['nodeX'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#posY', spec.menuValues['nodeY'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.levelerNode#posZ', spec.menuValues['nodeZ'])
		setXMLInt(spec.configXmlVehicle, 'vehicle.leveler.pickUpDirection', spec.menuValues['pickUpDirection'])
		setXMLString(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#effectNode', spec.effectNode)
		setXMLString(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#effectClass', spec.effectClass)
		setXMLString(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#materialType', spec.materialType)
		setXMLInt(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#materialTypeId', spec.materialTypeId)
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#fadeTime', spec.fadeTime)
		setXMLInt(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#speed', spec.speed)
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#maxHeight', spec.maxHeight)
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#posX', spec.menuValues['posX'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#posY', spec.menuValues['posY'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#posZ', spec.menuValues['posZ'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#scaleX', spec.menuValues['scaleX'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#scaleY', spec.menuValues['scaleY'])
		setXMLFloat(spec.configXmlVehicle, 'vehicle.leveler.effects.effectNode#scaleZ', spec.menuValues['scaleZ'])
		
		setXMLBool(spec.configXmlVehicle, 'vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration.fillUnits#allowFoldingWhileFilled', spec.allowFoldingWhileFilled)
		setXMLString(spec.configXmlVehicle, 'vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration.fillUnits.fillUnit#unit', spec.unit)
		setXMLString(spec.configXmlVehicle, 'vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration.fillUnits.fillUnit#fillTypeCategories', spec.fillTypeCategories)
		setXMLFloat(spec.configXmlVehicle, 'vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration.fillUnits.fillUnit#capacity', spec.menuValues['capacity'])
		setXMLBool(spec.configXmlVehicle, 'vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration.fillUnits.fillUnit#showInShop', spec.showInShop)
		
		setXMLBool(spec.configXmlVehicle, 'vehicle.leveler#active', spec.menuValues['modActive'])
		
		saveXMLFile(spec.configXmlVehicle)
		if reload then
			self:loadFromXmlRealShovelVehicle(self:getFullName())
		end
	end
end
function RealShovel:updateHeapEffect(index, value, noEventSend)
	local spec = self.spec_realShovel
	if not spec.modInitialized then
		return
	end
	
	if index ~= nil and value ~= nil then
		spec.menuValues[spec.menuItems[index][1]] = value
	end
	setTranslation(spec.effectNodeObject, spec.menuValues['posX'], spec.menuValues['posY'], spec.menuValues['posZ'])
	setScale(spec.effectNodeObject, spec.menuValues['scaleX'], spec.menuValues['scaleY'], spec.menuValues['scaleZ'])
	setTranslation(spec.nodes[1].node, spec.menuValues['nodeX'],spec.menuValues['nodeY'],spec.menuValues['nodeZ'])
	
	spec.nodes[1].width = spec.menuValues['width']
	spec.nodes[1].minDropWidth = spec.menuValues['minDropWidth']
	spec.nodes[1].maxDropWidth = spec.menuValues['maxDropWidth']
	spec.nodes[1].minDropHeight = spec.menuValues['minDropHeight']
	spec.nodes[1].maxDropHeight = spec.menuValues['maxDropHeight']
	spec.nodes[1].minDropDirOffset = spec.menuValues['minDropDirOffset']
	spec.nodes[1].maxDropDirOffset = spec.menuValues['maxDropDirOffset']
	spec.pickUpDirection = spec.menuValues['pickUpDirection']
	self.spec_fillUnit.fillUnits[spec.fillUnitIndex].capacity = spec.menuValues['capacity']
	
	self:saveToXmlRealShovelVehicle(false)
	
	RealShovelUpdateLevelerEvent.sendEvent(self, index, value, noEventSend)
end
function RealShovel:loadDefaultValues(update, noEventSend)
	local spec = self.spec_realShovel
	if not spec.modInitialized then
		return
	end
	
	local shovelWidth = self.spec_shovel.shovelNodes[1].width
	local shovelCapacity = self:getFillUnitCapacity(1)
	local x,rootY,z = getWorldRotation(self.rootNode)
	local x,shovelY,z = getWorldRotation(self.spec_shovel.shovelNodes[1].node)
	local dir = rootY * shovelY >= 0 and 1 or -1
	
	spec.menuItems = {	{'posX',			0},
						{'posY',			0.1},
						{'posZ',			-0.95},
						{'scaleX',			shovelWidth / 2.8},
						{'scaleY',			1},
						{'scaleZ',			1},
						{'width',			shovelWidth / 1.5},
						{'minDropWidth',	shovelWidth / 1.5 * 0.5},
						{'maxDropWidth',	shovelWidth / 1.5},
						{'minDropHeight',	0.02},
						{'maxDropHeight',	1.167},
						{'minDropDirOffset',0.6},
						{'maxDropDirOffset',0.6},
						{'pickUpDirection',	dir},
						{'capacity',		shovelCapacity * 1},
						{'nodeX',			0},
						{'nodeY',			0},
						{'nodeZ',			0.1},
						{'modActive',		true}}
	spec.menuValues = {}
	for _,item in pairs(spec.menuItems) do
		spec.menuValues[item[1]] = item[2]
	end
	if update then
		self:updateHeapEffect(nil, nil, true)
	end
	RealShovelLoadDefaultValuesEvent.sendEvent(self, noEventSend)
end
function RealShovel:onUpdate(dt, vehicle)
	local spec = self.spec_realShovel
	
	if not spec.modInitialized then
		return
	end
	if spec.firstRun then
		spec.firstRun = false
		local x,rootY,z = getWorldRotation(self.rootNode)
		local x,shovelY,z = getWorldRotation(self.spec_shovel.shovelNodes[1].node)
		local dir = rootY * shovelY >= 0 and 1 or -1
		self:updateHeapEffect(14, dir, true)
	end
	if self:getIsActive() then
		if spec.inputPlus or spec.inputMinus then
			if spec.selectedValue ~= 19 then
				setVisibility(getChildAt(spec.effectNodeObject, 2), false)
				setVisibility(getChildAt(spec.nodes[1].node, 0), false)
				local source = spec.effectNodeObject
				local index = spec.menuItems[spec.selectedValue][1]
				local val = spec.menuValues[index]
				local d = (spec.selectedValue == 15 and 10 or 0.01)
				spec.timerHud = spec.timerHud + dt
				spec.timerSettings = 0
				if spec.inputPlus then
					if spec.selectedValue == 14 then
						if spec.timerHud < dt + 1 then
							val = val * -1
						end
					else
						if spec.timerHud > 1000 then
							val = val + d
						else
							val = val + d / 5
						end
					end
				end
				if spec.inputMinus then
					if spec.selectedValue == 14 then
						if spec.timerHud == dt then
							val = val * -1
						end
					else
						if spec.timerHud > 1000 then
							val = val - d
						else
							val = val - d / 5
						end
					end
				end
				if spec.selectedValue == 15 then
					if val > 32000 or val < -32000 then
						val = 32000
					elseif val < 0 then
						val = 0
					end
				end
				spec.menuValues[index] = val
				setTranslation(source, spec.menuValues['posX'],spec.menuValues['posY'],spec.menuValues['posZ'])
				setScale(source, spec.menuValues['scaleX'],spec.menuValues['scaleY'],spec.menuValues['scaleZ'])
				setTranslation(spec.nodes[1].node, spec.menuValues['nodeX'],spec.menuValues['nodeY'],spec.menuValues['nodeZ'])
				if spec.selectedValue == 1 or spec.selectedValue == 2 or spec.selectedValue == 3 or spec.selectedValue == 4 or spec.selectedValue == 5 or spec.selectedValue == 6 then
					setVisibility(getChildAt(spec.effectNodeObject, 2), self:getFillUnitFillLevel(spec.fillUnitIndex) / self:getFillUnitCapacity(spec.fillUnitIndex) < 0.4 and true or false)
				elseif spec.selectedValue == 16 or spec.selectedValue == 17 or spec.selectedValue == 18 then
					setVisibility(getChildAt(spec.nodes[1].node, 0), true)
				end
			end
		else
			spec.timerHud = 0
			spec.timerSettings = spec.timerSettings + dt
			if spec.timerSettings > 1000 then
				spec.timerSettings = 0
				setVisibility(getChildAt(spec.effectNodeObject, 2), false)
				setVisibility(getChildAt(spec.nodes[1].node, 0), false)
			end
		end
	else
		setVisibility(getChildAt(spec.effectNodeObject, 2), false)
		setVisibility(getChildAt(spec.nodes[1].node, 0), false)
	end
    if self.isClient then
        local fillType = self:getFillUnitFillType(spec.fillUnitIndex)
        local fillLevel = self:getFillUnitFillLevel(spec.fillUnitIndex)
        local visible = fillLevel > 2*g_densityMapHeightManager:getMinValidLiterValue(fillType)
        if visible and fillType ~= FillType.UNKNOWN then
            g_effectManager:setFillType(spec.effects, fillType)
            g_effectManager:startEffects(spec.effects)
            local fillPercentage = fillLevel / self:getFillUnitCapacity(spec.fillUnitIndex)
            for _, effect in pairs(spec.effects) do
                effect:setFillLevel(fillPercentage)
                effect:setLastVehicleSpeed(self.movingDirection * self:getLastSpeed())
            end
        else
            g_effectManager:stopEffects(spec.effects)
        end
    end
    if self.isServer then
        for _, levelerNode in pairs(spec.nodes) do
            local x0,y0,z0 = localToWorld(levelerNode.node, -levelerNode.width, 0, levelerNode.maxDropDirOffset)
            local x1,y1,z1 = localToWorld(levelerNode.node,  levelerNode.width, 0, levelerNode.maxDropDirOffset)
            if not g_farmlandManager:getIsOwnedByFarmAtWorldPosition(self:getOwnerFarmId(), x0, z0) or
               not g_farmlandManager:getIsOwnedByFarmAtWorldPosition(self:getOwnerFarmId(), x1, z1) then
                break
            end
            local pickedUpFillLevel = 0
            local fillType = self:getFillUnitFillType(spec.fillUnitIndex)
            local fillLevel = self:getFillUnitFillLevel(spec.fillUnitIndex)
            if fillType == FillType.UNKNOWN or fillLevel < g_densityMapHeightManager:getMinValidLiterValue(fillType) + 0.001 then
				
                local newFillType = DensityMapHeightUtil.getFillTypeAtLine(x0,y0,z0, x1,y1,z1, 0.5*levelerNode.maxDropDirOffset)
                if newFillType ~= FillType.UNKNOWN and newFillType ~= fillType then
                    self:addFillUnitFillLevel(self:getOwnerFarmId(), spec.fillUnitIndex, -math.huge)
                    fillType = newFillType
                end
            end
            local heightType = g_densityMapHeightManager:getDensityMapHeightTypeByFillTypeIndex(fillType)
            if fillType ~= FillType.UNKNOWN and heightType ~= nil then
                local innerRadius = 0
                local outerRadius = DensityMapHeightUtil.getDefaultMaxRadius(fillType)
                local capacity = self:getFillUnitCapacity(spec.fillUnitIndex)
                -- pick up at node
                if self:getIsLevelerPickupNodeActive(levelerNode) then
                    if spec.pickUpDirection == self.movingDirection then
                        local sx,sy,sz = localToWorld(levelerNode.node, -levelerNode.width, 0, 0)
                        local ex,ey,ez = localToWorld(levelerNode.node,  levelerNode.width, 0, 0)
                        fillLevel = self:getFillUnitFillLevel(spec.fillUnitIndex)
                        local delta = -(capacity-fillLevel)
                        local numHeightLimitChecks = levelerNode.numHeightLimitChecks
                        if numHeightLimitChecks > 0 then
                            local movementY = 0
                            for i=0,numHeightLimitChecks do
                                local t = i/numHeightLimitChecks
                                local xi = sx + (ex-sx)*t
                                local yi = sy + (ey-sy)*t
                                local zi = sz + (ez-sz)*t
                                local hi = DensityMapHeightUtil.getHeightAtWorldPos(xi,yi,zi)
                                movementY = math.max(movementY, hi-0.05 - yi) -- limit to 5cm below surface
                            end
                            if movementY > 0 then
                                sy = sy+movementY
                                ey = ey+movementY
                            end
                        end
                        levelerNode.lastPickUp, levelerNode.lineOffsetPickUp = DensityMapHeightUtil.tipToGroundAroundLine(self, delta, fillType, sx,sy-0.1,sz, ex,ey-0.1,ez, innerRadius, outerRadius, levelerNode.lineOffsetPickUp, true, nil)
                        if levelerNode.lastPickUp < 0 then
                            levelerNode.lastPickUp = levelerNode.lastPickUp + spec.litersToPickup
                            spec.litersToPickup = 0
                            self:addFillUnitFillLevel(self:getOwnerFarmId(), spec.fillUnitIndex, -levelerNode.lastPickUp, fillType, ToolType.UNDEFINED, nil)
                            pickedUpFillLevel = levelerNode.lastPickUp
                        end
                    end
                end
                -- drop at node
                fillLevel = self:getFillUnitFillLevel(spec.fillUnitIndex)
                if fillLevel > 0 then
                    local f = (fillLevel/capacity)
                    local width = MathUtil.lerp(levelerNode.minDropWidth, levelerNode.maxDropWidth, f)
                    local sx,sy,sz = localToWorld(levelerNode.node, -width, 0, 0)
                    local ex,ey,ez = localToWorld(levelerNode.node,  width, 0, 0)
                    local yOffset = -0.1 -0.05
                    levelerNode.lastDrop1, levelerNode.lineOffsetDrop1 = DensityMapHeightUtil.tipToGroundAroundLine(self, fillLevel, fillType, sx,sy+yOffset,sz, ex,ey+yOffset,ez, innerRadius, outerRadius, levelerNode.lineOffsetDrop1, true, tipOcclusionAreas)
					if levelerNode.lastDrop1 > 0 then
                        local leftOver = fillLevel - levelerNode.lastDrop1
                        if leftOver <= g_densityMapHeightManager:getMinValidLiterValue(fillType) then
                            levelerNode.lastDrop1 = fillLevel
                            spec.litersToPickup = spec.litersToPickup + leftOver
                        end
                        self:addFillUnitFillLevel(self:getOwnerFarmId(), spec.fillUnitIndex, -levelerNode.lastDrop1, fillType, ToolType.UNDEFINED, nil)
                    end
				end
                -- drop further at front
                fillLevel = self:getFillUnitFillLevel(spec.fillUnitIndex)
                if fillLevel > 0 then
                    local f = (fillLevel/capacity)
                    local width = MathUtil.lerp(levelerNode.minDropWidth, levelerNode.maxDropWidth, f)
                    local yOffset = MathUtil.lerp(levelerNode.minDropHeight, levelerNode.maxDropHeight, f)
                    local sx,sy,sz = localToWorld(levelerNode.node, -width, 0, 0)
                    local ex,ey,ez = localToWorld(levelerNode.node,  width, 0, 0)
                    local dx,dy,dz = localDirectionToWorld(levelerNode.node, 0, 0, 1)
                    local backOffset = -outerRadius * math.abs(spec.pickUpDirection) * 1.5
                    local backLen = MathUtil.lerp(levelerNode.minDropDirOffset, levelerNode.maxDropDirOffset, f) - backOffset
                    local backX,backY,backZ = dx*backOffset,dy*backOffset,dz*backOffset
                    dx,dy,dz = dx*backLen,dy*backLen,dz*backLen
                    local terrainHeightUpdater = g_densityMapHeightManager:getTerrainDetailHeightUpdater()
                    if terrainHeightUpdater ~= nil then
                        addDensityMapHeightOcclusionArea(terrainHeightUpdater, sx+backX,sy+backY,sz+backZ, ex-sx,ey-sy,ez-sz, dx, dy, dz, true)
                        if width < levelerNode.width-0.05 then
                            -- fully block left and right of the inner block area
                            local sx2,sy2,sz2 = localToWorld(levelerNode.node, -levelerNode.width, 0, 0)
                            local ex2,ey2,ez2 = localToWorld(levelerNode.node,  levelerNode.width, 0, 0)
                            addDensityMapHeightOcclusionArea(terrainHeightUpdater, sx2+backX,sy2+backY,sz2+backZ, sx-sx2,sy-sy2,sz-sz2, dx, dy, dz, false)
                            addDensityMapHeightOcclusionArea(terrainHeightUpdater, ex +backX,ey +backY,ez +backZ, ex2-ex,ey2-ey,ez2-ez, dx, dy, dz, false)
                        end
                    end
                    levelerNode.lastDrop2, levelerNode.lineOffsetDrop2 = DensityMapHeightUtil.tipToGroundAroundLine(self, fillLevel, fillType, sx,sy+yOffset,sz, ex,ey+yOffset,ez, 0, outerRadius, levelerNode.lineOffsetDrop2, true, nil)
					if levelerNode.lastDrop2 > 0 then
                        local leftOver = fillLevel - levelerNode.lastDrop2
                        if leftOver <= g_densityMapHeightManager:getMinValidLiterValue(fillType) then
                            levelerNode.lastDrop2 = fillLevel
                            spec.litersToPickup = spec.litersToPickup + leftOver
                        end
                        self:addFillUnitFillLevel(self:getOwnerFarmId(), spec.fillUnitIndex, -levelerNode.lastDrop2, fillType, ToolType.UNDEFINED, nil)
                    end
                end
            end
            -- call fill level changed callack to inform bunker silo about change
            if pickedUpFillLevel < 0 and fillType ~= FillType.UNKNOWN then
                self:notifiyBunkerSilo(pickedUpFillLevel, fillType)
            end
        end
    end
	if self.isClient then
		if self:getIsActive() then
			if spec.event_IDs ~= nil and g_dedicatedServerInfo == nil then
				g_inputBinding:setActionEventText(spec.event_IDs['REALSHOVEL_MENU_TOGGLE'], spec.showHud and g_i18n:getText('REALSHOVEL_MENU_HIDE') or g_i18n:getText('REALSHOVEL_MENU_SHOW'))
				g_inputBinding:setActionEventText(spec.event_IDs['REALSHOVEL_TOGGLE_HELP'], RealShovel.showHelp and g_i18n:getText('REALSHOVEL_HIDE_HELP') or g_i18n:getText('REALSHOVEL_SHOW_HELP'))
				for actionName,eventID in pairs(spec.event_IDs) do
					if actionName == InputAction.REALSHOVEL_TOGGLE_HELP then
						g_inputBinding:setActionEventTextVisibility(eventID, true)
					elseif actionName == InputAction.REALSHOVEL_TOGGLE then
						g_inputBinding:setActionEventTextVisibility(eventID, false)
					else
						g_inputBinding:setActionEventTextVisibility(eventID, RealShovel.showHelp)
					end
				
					if actionName == InputAction.REALSHOVEL_TOGGLE then
						g_inputBinding:setActionEventActive(eventID, true)
					elseif actionName == InputAction.REALSHOVEL_MENU_PLUS or actionName == InputAction.REALSHOVEL_MENU_MINUS or actionName == InputAction.REALSHOVEL_MENU_CHANGE_DOWN or actionName == InputAction.REALSHOVEL_MENU_CHANGE_UP then
						g_inputBinding:setActionEventActive(eventID, spec.showHud and spec.menuValues['modActive'])
					else
						g_inputBinding:setActionEventActive(eventID, spec.menuValues['modActive'])
					end
				end
			end
		end
	end
end
function RealShovel:onDraw()
	local spec = self.spec_realShovel
	
	if not spec.modInitialized then
		return
	end
	
	if self.isClient and self:getIsActiveForInput() and spec.showHud and spec.menuValues['modActive'] then
		renderOverlay(spec.settingHud, RealShovel.hudPositionX, RealShovel.hudPositionY, spec.hudWidth, spec.hudHeight)
		for i = 1, #spec.menuItems-1, 1 do
			if i == 1 then
				spec.hudText1 = spec.hudText1 .. '\n'
				spec.hudText2 = spec.hudText2 .. 'Heap Effect\n'
				spec.hudText3 = spec.hudText3 .. '\n'
			elseif i == 7 then
				spec.hudText1 = spec.hudText1 .. '\n\n'
				spec.hudText2 = spec.hudText2 .. '-----------------------------------\nLeveler\n'
				spec.hudText3 = spec.hudText3 .. '\n\n'
			end
			if i == spec.selectedValue then
				spec.hudText1 = spec.hudText1 .. 'X\n'
			else
				spec.hudText1 = spec.hudText1 .. '\n'
			end
			spec.hudText2 = spec.hudText2 .. spec.menuItems[i][1]..'\n'
			if i == 15 then
				spec.hudText3 = spec.hudText3 .. string.format('%.0f',spec.menuValues[spec.menuItems[i][1]])..'\n'
			else
				spec.hudText3 = spec.hudText3 .. string.format('%.2f',spec.menuValues[spec.menuItems[i][1]])..'\n'
			end
		end
		spec.hudText1 = spec.selectedValue == 19 and  spec.hudText1 .. '\nX\n' or spec.hudText1 .. '\n\n'
		spec.hudText2 = spec.hudText2 .. '------------------------------------\nReset to default\n'
		spec.hudText3 = spec.hudText3 .. '\n\n'
		setTextBold(false);
		setTextColor(1, 1, 1, 1)
		setTextAlignment(RenderText.ALIGN_LEFT)
		renderText(spec.hudText1X, spec.hudTextY, 0.016*spec.sizeHud, spec.hudText1)
		renderText(spec.hudText2X, spec.hudTextY, 0.016*spec.sizeHud, spec.hudText2)
		renderText(spec.hudText3X, spec.hudTextY, 0.016*spec.sizeHud, spec.hudText3)
		spec.hudText1 = ''
		spec.hudText2 = ''
		spec.hudText3 = ''
	end
end
function RealShovel:actionCallback(actionName, keyStatus, arg4, arg5, arg6)
	local spec = self.spec_realShovel
	
	if not spec.modInitialized then
		return
	end
	
	if keyStatus > 0 then
		if actionName == 'REALSHOVEL_MENU_CHANGE_DOWN' then
			if spec.selectedValue + 1 <= #spec.menuItems - 0 then
				spec.selectedValue = spec.selectedValue + 1
			else
				spec.selectedValue = 1
			end
		elseif actionName == 'REALSHOVEL_MENU_CHANGE_UP' then
			if spec.selectedValue - 1 >= 1 then
				spec.selectedValue = spec.selectedValue - 1
			else
				spec.selectedValue = #spec.menuItems - 0
			end
		elseif actionName == 'REALSHOVEL_MENU_TOGGLE' then
			spec.showHud = not spec.showHud
			
		elseif actionName == 'REALSHOVEL_TOGGLE_HELP' then
			RealShovel.showHelp = not RealShovel.showHelp
			self:saveToXmlRealShovel()
		elseif actionName == 'REALSHOVEL_TOGGLE' then
			spec.menuValues['modActive'] = not spec.menuValues['modActive']
			self:activateModRS(spec.menuValues['modActive'], 'THIS')
			self:updateHeapEffect(19, spec.menuValues['modActive'], false) 
		end
	end
	if actionName == 'REALSHOVEL_MENU_PLUS' then
		spec.inputPlus = keyStatus > 0
		if keyStatus == 0 then
			if spec.selectedValue == 19 then
				self:loadDefaultValues(true, false)
			else
				self:updateHeapEffect(spec.selectedValue, spec.menuValues[spec.menuItems[spec.selectedValue][1]], false)
			end
		end
	elseif actionName == 'REALSHOVEL_MENU_MINUS' then
		spec.inputMinus = keyStatus > 0
		if keyStatus == 0 then
			if spec.selectedValue == 19 then
				self:loadDefaultValues(true, false)
			else
				self:updateHeapEffect(spec.selectedValue, spec.menuValues[spec.menuItems[spec.selectedValue][1]], false)
			end
		end
	end
end
function RealShovel:activateModRS(status, mode)
	local spec = self.spec_realShovel
	
	if mode == 'THIS' then
		mode = self:getFullName()
	end
	local text = 'ACTIVATED'
	if not status then
		text = 'DEACTIVATED'
	end
	g_currentMission:showBlinkingWarning("REAL SHOVEL FUNCTIONS "..tostring(text).." FOR "..tostring(mode).." VEHICLES", 2000);
end
function RealShovel:onReadStream(streamId, connection)
	local spec = self.spec_realShovel
	if not spec.modInitialized then
		return
	end
	
	local posX = streamReadFloat32(streamId)
	local posY = streamReadFloat32(streamId)
	local posZ = streamReadFloat32(streamId)
	local scaleX = streamReadFloat32(streamId)
	local scaleY = streamReadFloat32(streamId)
	local scaleZ = streamReadFloat32(streamId)
	local width = streamReadFloat32(streamId)
	local minDropWidth = streamReadFloat32(streamId)
	local maxDropWidth = streamReadFloat32(streamId)
	local minDropHeight = streamReadFloat32(streamId)
	local maxDropHeight = streamReadFloat32(streamId)
	local minDropDirOffset = streamReadFloat32(streamId)
	local maxDropDirOffset = streamReadFloat32(streamId)
	local pickUpDirection = streamReadInt8(streamId)
	local capacity = streamReadInt16(streamId)
	local nodeX = streamReadFloat32(streamId)
	local nodeY = streamReadFloat32(streamId)
	local nodeZ = streamReadFloat32(streamId)
	local modActive = streamReadBool(streamId)
	
	if posX ~= nil and posY ~= nil and posZ ~= nil and scaleX ~= nil and scaleY ~= nil and scaleZ ~= nil and width ~= nil and minDropWidth ~= nil and maxDropWidth ~= nil and minDropHeight ~= nil and maxDropHeight ~= nil
		and minDropDirOffset ~= nil and maxDropDirOffset ~= nil and pickUpDirection ~= nil and capacity ~= nil and nodeX ~= nil and nodeY ~= nil and nodeZ ~= nil and modActive ~= nil then
		spec.menuValues['posX'] = posX
		spec.menuValues['posY'] = posY
		spec.menuValues['posZ'] = posZ
		spec.menuValues['scaleX'] = scaleX
		spec.menuValues['scaleY'] = scaleY
		spec.menuValues['scaleZ'] = scaleZ
		spec.menuValues['width'] = width
		spec.menuValues['minDropWidth'] = minDropWidth
		spec.menuValues['maxDropWidth'] = maxDropWidth
		spec.menuValues['minDropHeight'] = minDropHeight
		spec.menuValues['maxDropHeight'] = maxDropHeight
		spec.menuValues['minDropDirOffset'] = minDropDirOffset
		spec.menuValues['maxDropDirOffset'] = maxDropDirOffset
		spec.menuValues['pickUpDirection'] = pickUpDirection
		spec.menuValues['capacity'] = capacity
		spec.menuValues['nodeX'] = nodeX
		spec.menuValues['nodeY'] = nodeY
		spec.menuValues['nodeZ'] = nodeZ
		spec.menuValues['modActive'] = modActive
		self:updateHeapEffect(nil, nil, true)
	end
end;
function RealShovel:onWriteStream(streamId, connection)
	local spec = self.spec_realShovel
	if not spec.modInitialized then
		return
	end
	
	streamWriteFloat32(streamId, spec.menuValues['posX'])
	streamWriteFloat32(streamId, spec.menuValues['posY'])
	streamWriteFloat32(streamId, spec.menuValues['posZ'])
	streamWriteFloat32(streamId, spec.menuValues['scaleX'])
	streamWriteFloat32(streamId, spec.menuValues['scaleY'])
	streamWriteFloat32(streamId, spec.menuValues['scaleZ'])
	streamWriteFloat32(streamId, spec.menuValues['width'])
	streamWriteFloat32(streamId, spec.menuValues['minDropWidth'])
	streamWriteFloat32(streamId, spec.menuValues['maxDropWidth'])
	streamWriteFloat32(streamId, spec.menuValues['minDropHeight'])
	streamWriteFloat32(streamId, spec.menuValues['maxDropHeight'])
	streamWriteFloat32(streamId, spec.menuValues['minDropDirOffset'])
	streamWriteFloat32(streamId, spec.menuValues['maxDropDirOffset'])
	streamWriteInt8(streamId, spec.menuValues['pickUpDirection'])
	streamWriteInt16(streamId, spec.menuValues['capacity'])
	streamWriteFloat32(streamId, spec.menuValues['nodeX'])
	streamWriteFloat32(streamId, spec.menuValues['nodeY'])
	streamWriteFloat32(streamId, spec.menuValues['nodeZ'])
	streamWriteBool(streamId, spec.menuValues['modActive'])
end

--- Events ---
RealShovelUpdateLevelerEvent = {}
RealShovelUpdateLevelerEvent_mt = Class(RealShovelUpdateLevelerEvent, Event)

InitEventClass(RealShovelUpdateLevelerEvent, "RealShovelUpdateLevelerEvent")

function RealShovelUpdateLevelerEvent:emptyNew()
    local self = Event:new(RealShovelUpdateLevelerEvent_mt)
    return self
end
    
function RealShovelUpdateLevelerEvent:new(object, index, value)
	local self = RealShovelUpdateLevelerEvent:emptyNew()
	self.object = object
	self.index = index
	self.value = value
	return self
end

function RealShovelUpdateLevelerEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
	self.index = streamReadInt8(streamId)
	if self.index == 14 or self.index == 15 then
		self.value = streamReadInt16(streamId)
	elseif self.index == 19 then
		self.value = streamReadBool(streamId)
	else
		self.value = streamReadFloat32(streamId)
	end
	self:run(connection)
end

function RealShovelUpdateLevelerEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteInt8(streamId, self.index)
	if self.index == 14 or self.index == 15 then
		streamWriteInt16(streamId, self.value)
	elseif self.index == 19 then
		streamWriteBool(streamId, self.value)
	else
		streamWriteFloat32(streamId, self.value)
	end
end

function RealShovelUpdateLevelerEvent:run(connection)
	if not connection:getIsServer() then
		g_server:broadcastEvent(self, false, connection, self.object)
	end;
	if self.object ~= nil then
		self.object:updateHeapEffect(self.index, self.value, true)
	end
end

function RealShovelUpdateLevelerEvent.sendEvent(vehicle, index, value, noEventSend)
	if index ~= vehicle.index and value ~= vehicle.value then
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(RealShovelUpdateLevelerEvent:new(vehicle, index, value), nil, nil, vehicle)
			else
				g_client:getServerConnection():sendEvent(RealShovelUpdateLevelerEvent:new(vehicle, index, value))
			end
		end
	end
end

RealShovelLoadDefaultValuesEvent = {}
RealShovelLoadDefaultValuesEvent_mt = Class(RealShovelLoadDefaultValuesEvent, Event)

InitEventClass(RealShovelLoadDefaultValuesEvent, "RealShovelLoadDefaultValuesEvent")

function RealShovelLoadDefaultValuesEvent:emptyNew()
    local self = Event:new(RealShovelLoadDefaultValuesEvent_mt)
    return self
end
    
function RealShovelLoadDefaultValuesEvent:new(object)
	local self = RealShovelLoadDefaultValuesEvent:emptyNew()
	self.object = object
	return self
end

function RealShovelLoadDefaultValuesEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
	self:run(connection)
end

function RealShovelLoadDefaultValuesEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
end

function RealShovelLoadDefaultValuesEvent:run(connection)
	if not connection:getIsServer() then
		g_server:broadcastEvent(self, false, connection, self.object)
	end;
	if self.object ~= nil then
		self.object:loadDefaultValues(true, true)
	end
end

function RealShovelLoadDefaultValuesEvent.sendEvent(vehicle, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(RealShovelLoadDefaultValuesEvent:new(vehicle), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(RealShovelLoadDefaultValuesEvent:new(vehicle))
		end
	end
end